/*	File			BullyElectionAlgorithm.java
	Purpose			Implementation of the Bully Election Algorithm
	Author			Richard Anthony	(R.J.Anthony@gre.ac.uk)
	Date			March 2015

	Demonstration implementation of the Bully election algorithm.
	Normal process states in this algorithm are State_Slave and State_Master.
	
	All processes are initialised in State_Slave. In this state they listen for MT_MasterHeartbeat messages,
	sent periodically by a master (if one exists). If one or more slaves miss a sequence of the heartbeat 
	messages (indicating absence of a master) they initiate an election by sending a MT_ElectionTrigger message.

	The process that wins the election will send MT_MasterHeartbeat messages periodically. A very important
	rule is that there must only be one State_Master instance present at any time, and therefore the election
	mechanism must be robust to prevent multiple masters locally determining that they have won an election.
	To facilitate this a process initially enters the State_Candidate state during an election. This is a 
	transient state used to ensure that all challenges have been resolved before a single process elevates to
	State_Master.

	The following design assumptions are made:
	1.	Elections will occur between processes in a local network and therefore broadcast communication can be used.
	2.	Only one participating process will be present on each computer, so the IP address of the host computer can be
		used as the process’ ID for the purposes of the election (this is a simple way to ensure that IDs are unique).
	3.	The process with the highest ID (its IP address) wins an election.
	4.	Once a single master exists, slaves with higher IDs may join the system (without triggering an election).
*/
package BullyElectionAlgorithm;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.*;
import java.util.Arrays;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Random;
import javax.swing.ScrollPaneConstants;
import javax.swing.*;

public class BullyElectionAlgorithm extends javax.swing.JFrame implements ActionListener
{
    private static int BULLY_ELECTION_ALGORITHM_PORT = 8005;	// All messages are received ion this port, regardless of the processes election state
    private static int MASTER_HEARTBEAT_TX_PERIOD = 5000;       // Heartbeat transmission period (milliseconds)
    private static int ELECTION_TRIGGER_HEARTBEATS_MISSED = 3;  // Threshold number of heartbeats a Slave must miss before initiating an election
    private static int CHECK_FOR_MESSAGE_PERIOD = 100;          // Period between checking for any incomming messages (milliseconds)
    private static int CANDIDATE_STATE_DURATION = 3;            // The number of heartbeat periods a node must wait in Candidate state before promoting
    private static int HeartbeatClock_LIMIT = 1000;
    private static int TIMER_MASTER_HEARTBEAT_TX = 1;
    private static int TIMER_CHECK_FOR_MESSAGE = 2;

    public enum BULLY_ELECTION_ALGORITHM_Process_State {State_Slave, State_Candidate, State_Master};
    public enum BULLY_ELECTION_ALGORITHM_Message_Type {MT_MasterHeartbeat, MT_ElectionTrigger};

    // To implement the equivallent of a C/C++/C# structure, in Java a class is used
    final class BULLY_ELECTION_ALGORITHM_Message
    {
        BULLY_ELECTION_ALGORITHM_Message_Type eMessageType;   //size 4 Offset 0
    };

    private static int m_iBULLY_ELECTION_ALGORITHM_Message_Size = 4;

    BULLY_ELECTION_ALGORITHM_Process_State m_LocalState;

    DatagramSocket m_RecvSock;
    MulticastSocket m_SendSock;
    InetAddress m_Local_IPAddress;      // Local Address for receiving all messages
    InetAddress m_Broadcast_IPAddress;  // Address for sending all messages (all are broadcast)
    Timer m_MasterHeartbeat_Timer;
    Timer m_MessageReceive_Timer;

    int m_iMasterHeartbeatClock;	// Used to measure time intervals between heartbeats from the Master instance
    int m_iMostRecentHeartbeat_TimeDetected;
    int m_iClock_CountInCandidateState_SincelastChallenge; // Keeps track of how long the node has been in Candidate state
						    // Must stay in candidate state long enough to ensure all possible challengers have been defeated
    DefaultListModel m_listModel_EventLog; // For use with JList

    public BullyElectionAlgorithm()
    {
        m_listModel_EventLog = new DefaultListModel();
        initComponents();
        m_LocalState = BULLY_ELECTION_ALGORITHM_Process_State.State_Slave;
        m_iMasterHeartbeatClock = 0;	// Used to measure time intervals between heartbeats from the Master instance
        m_iMostRecentHeartbeat_TimeDetected = m_iMasterHeartbeatClock;
        m_iClock_CountInCandidateState_SincelastChallenge = 0;
 
        try
        {
           m_Local_IPAddress = InetAddress.getLocalHost();
           m_Broadcast_IPAddress = GetLocalBroadcastAddress_From_LocalHostAddress(m_Local_IPAddress);
        }   
        catch (Exception Ex) 
        {             
           JOptionPane.showMessageDialog(null,"Could not get local address or broadcast address ... Exiting","Event Notification - Publish Client (Java)",0);
           System.exit(-1);
        }
        
        jTextField_CurrentState.setText("");
        jButton_ClearLog.setEnabled(true);
        jButton_Done.setEnabled(true);
        jList_EventLog.setEnabled(false);
        CreateSendSocket();             // The socket used for all sends
        CreateReceiveSocket();          // The socket on which all messages are received
        Start_MasterHeartbeatTimer();
        Start_MessageReceiveTimer();
        DisplayLocalIPAddress();
        DisplayLocalState();
        WriteEventLogEntry("Initialised (Slave state)");
    }

    void DisplayLocalIPAddress()
    {   
        String sHostAddress = "";
        try
        {
            sHostAddress = m_Local_IPAddress.toString(); // Expected format: "Hostname/192.168.100.5"
            String sStr = "";
            String sHostName = "";
            int iStartOfDottedIPSubstring = sHostAddress.indexOf('/');
            if(-1 == iStartOfDottedIPSubstring)
            {   // Align to the actual start of the dotted IP address part of the string
                iStartOfDottedIPSubstring = 0; // '/' not found
            }
            else
            {
                sHostName = sHostAddress.substring(0, iStartOfDottedIPSubstring);
                iStartOfDottedIPSubstring++; // Skip the '/' charachter
            }
            String sDottedIPAddress = sHostAddress.substring(iStartOfDottedIPSubstring);
            sStr = String.format("Local host:%s  IPaddress:%s", sHostName, sDottedIPAddress);
            WriteEventLogEntry(sStr);
        }
        catch (Exception Ex)
        {
        }
    }
    
   void CreateSendSocket()
    {
        try
        {
            m_SendSock = new MulticastSocket();
            m_SendSock.setSoTimeout(100);   // Timeout  = 100ms (i.e. non-blocking IO behaviour)
            m_SendSock.setBroadcast(true);  // Set broadcast mode
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"Bully Election Algorithm",0);
           CloseSocketsAndExit();
        }
    }

    void CreateReceiveSocket()
    {   
        try
        {
            m_RecvSock = new DatagramSocket(BULLY_ELECTION_ALGORITHM_PORT); // create socket and bind to the port for receiveing
            m_RecvSock.setSoTimeout(100);   // Timeout  = 100ms (i.e. non-blocking IO behaviour)
        }
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"Bully Election Algorithm",0);
           CloseSocketsAndExit();
        }
    }

    void Start_MasterHeartbeatTimer()
    {	
        m_MasterHeartbeat_Timer = new Timer();
        m_MasterHeartbeat_Timer.scheduleAtFixedRate(
        new OnTimedEvent_MasterHeartbeatTimer(), MASTER_HEARTBEAT_TX_PERIOD, MASTER_HEARTBEAT_TX_PERIOD);
    }

    void Stop_MasterHeartbeatTimer()
    {
        if (null != m_MasterHeartbeat_Timer)
        {
            m_MasterHeartbeat_Timer.cancel();
        }
    }

    void Start_MessageReceiveTimer()
    {	
        m_MessageReceive_Timer = new Timer();
        m_MessageReceive_Timer.scheduleAtFixedRate(
        new OnTimedEvent_MessageReceiveTimer(), CHECK_FOR_MESSAGE_PERIOD, CHECK_FOR_MESSAGE_PERIOD);
    }

    void Stop_MessageReceiveTimer()
    {
        if (null != m_MessageReceive_Timer)
        {
            m_MessageReceive_Timer.cancel();
        }
    }

    class OnTimedEvent_MasterHeartbeatTimer extends TimerTask
    {
        public void run()
        {
            m_iMasterHeartbeatClock++;		// Used to measure time intervals between heartbeats from the Master (only relevent when in Slave state)
            m_iMasterHeartbeatClock %= HeartbeatClock_LIMIT;	// m_iHeartbeatClock counts from 0 to 999, then is reset to 0 and start counting again

            switch(m_LocalState)
            {
                case State_Master:
                    SendMasterHeartbeatMessage();	// Only performed if current state is State_Master	
                    break;
                case State_Candidate:
                    m_iClock_CountInCandidateState_SincelastChallenge++; // Increment Candidate duration clock
                    if(CANDIDATE_STATE_DURATION <= m_iClock_CountInCandidateState_SincelastChallenge)
                    {	// Have stayed in candidate state for the requisite period (no successful challengers, so now promote to Master)				
                        TransitionToMasterState();
                    }
                    break;
                case State_Slave:
                    CheckPresenceOfMaster(m_iMasterHeartbeatClock);		// Check master is still present based on interval since last heartbeat received
                    break;
            }
        }
    }

    class OnTimedEvent_MessageReceiveTimer extends TimerTask
    {
        public void run()
        {
            DoReceive();
        }
    }
        
    void CheckPresenceOfMaster(int iCurrentHeartbeatClock)
    {	// Check interval since last heartbeat to determine if a Master instance is still present	
        int iEffective_CurrentHeartbeatClock; // Takes care of clock rollovers

        // Heartbeat count increments every MASTER_HEARTBEAT_TX_PERIOD milliseconds, up to a maximum of HeartbeatClock_LIMIT increments (modulo)
        // Heartbeat count rolls over when it reaches HeartbeatClock_LIMIT, back to 0 (%HeartbeatClock_LIMIT)
        if(iCurrentHeartbeatClock < m_iMostRecentHeartbeat_TimeDetected)
        {   // Current clock value is lower, so it has rolled over! Treat it as having a value of +HeartbeatClock_LIMIT
            // for the purposes of the calculations (so the current clock value is always higher than the stored value
            // in iMostRecentHeartbeat_TimeDetected).
            iEffective_CurrentHeartbeatClock = iCurrentHeartbeatClock + HeartbeatClock_LIMIT;
        }
        else
        {   // Use Current clock value as is.
            iEffective_CurrentHeartbeatClock = iCurrentHeartbeatClock;
        }

        if((m_iMostRecentHeartbeat_TimeDetected + ELECTION_TRIGGER_HEARTBEATS_MISSED)
                <= iEffective_CurrentHeartbeatClock)
        {   // m_iMostRecentHeartbeat_TimeDetected value is out-of-date, Master is considered to have departed
            TransitionToCandidateState(); //Start election, with self as a candidate
        }
    }

    void TransitionToMasterState()
    {
        ChangeState(BULLY_ELECTION_ALGORITHM_Process_State.State_Master);
        SendMasterHeartbeatMessage();	// Immediately send a heartbeat message to prevent further slave elections
    }

    void TransitionToCandidateState()
    {
        ChangeState(BULLY_ELECTION_ALGORITHM_Process_State.State_Candidate);
        m_iClock_CountInCandidateState_SincelastChallenge = 0; // Reset Candidate duration clock
        SendElectionTriggerMessage();	//Start election, with self as a candidate
    }

    void TransitionToSlaveState()
    {
        ChangeState(BULLY_ELECTION_ALGORITHM_Process_State.State_Slave);
        m_iMostRecentHeartbeat_TimeDetected = m_iMasterHeartbeatClock; 	// Update MostRecentHeartbeat timestamp
    }

    byte[] Serialize_Message_PDU(BULLY_ELECTION_ALGORITHM_Message Message)
    {   // Convert an BULLY_ELECTION_ALGORITHM_Message structure into a byte array
        // Serialisation is very simple in this case, since there is only one Field in the message structure
        byte[] bByteArray = new byte[m_iBULLY_ELECTION_ALGORITHM_Message_Size]; // Create a byte array to hold the serialised form of the message PDU
        bByteArray[0] = (byte)Message.eMessageType.ordinal();
        return bByteArray;
    }
         
    void SendMessage(BULLY_ELECTION_ALGORITHM_Message Message)
    {
        byte[] bSendBuf = Serialize_Message_PDU(Message);
        try
        {
            DatagramPacket SendPacket = new DatagramPacket(bSendBuf, bSendBuf.length,
                                    m_Broadcast_IPAddress /*The address to send to*/, BULLY_ELECTION_ALGORITHM_PORT); 
            m_SendSock.send(SendPacket);
        }
        catch(SocketTimeoutException ex)
        {
             return;
        }
        catch(Exception ex)
        {
            System.out.printf("Send failed: %s\n", ex.toString());
            return;
        }
    }

    void SendMasterHeartbeatMessage()
    {
        BULLY_ELECTION_ALGORITHM_Message Message = new BULLY_ELECTION_ALGORITHM_Message();
        Message.eMessageType = BULLY_ELECTION_ALGORITHM_Message_Type.MT_MasterHeartbeat;
        SendMessage(Message);
        WriteEventLogEntry("MasterHeartbeatMessage sent");
    }

    void SendElectionTriggerMessage()
    {
        BULLY_ELECTION_ALGORITHM_Message Message = new BULLY_ELECTION_ALGORITHM_Message();
        Message.eMessageType = BULLY_ELECTION_ALGORITHM_Message_Type.MT_ElectionTrigger;
        SendMessage(Message);
        WriteEventLogEntry("ElectionTriggerMessage sent");
    }

    BULLY_ELECTION_ALGORITHM_Message DeSerialize_ReceivedMessage(byte[] bRecdByteArray)
    {   // Converts a byte array (from Receive) to a BULLY_ELECTION_ALGORITHM_Message structure (Fields have fixed length, but non-fixed data length)
        // Serialisation is very simple in this case, since there is only one Field in the message structure
        BULLY_ELECTION_ALGORITHM_Message result_Message = new BULLY_ELECTION_ALGORITHM_Message();
        result_Message.eMessageType = BULLY_ELECTION_ALGORITHM_Message_Type.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
        return result_Message;
    }
        
    void DoReceive()
    {
        byte[] bRecvBuf = new byte[m_iBULLY_ELECTION_ALGORITHM_Message_Size];
        DatagramPacket RecvPacket = new DatagramPacket(bRecvBuf, m_iBULLY_ELECTION_ALGORITHM_Message_Size);
        InetAddress SenderIPAddress;

        try
        {
            m_RecvSock.receive(RecvPacket);
            // Place the sender's address into the m_SenderIPAddress
            SenderIPAddress = RecvPacket.getAddress();
        }
        catch(SocketTimeoutException ex)
        {
             return;
        }
        catch(Exception ex)
        {
            System.out.printf("Receive failed: %s\n", ex.toString());
            return;
        }

        if (0 >= RecvPacket.getLength())
        {
            return;
        }

        if (true == CompareNodeIDs_SenderSameAsLocal(SenderIPAddress))
        {	// Ignore messages from self (very important when using broadcast)
            WriteEventLogEntry("Ignore message from self");
            return;
        }

        BULLY_ELECTION_ALGORITHM_Message Message = DeSerialize_ReceivedMessage(bRecvBuf);

        switch (Message.eMessageType)
        {
            case MT_MasterHeartbeat:
                WriteEventLogEntry("MasterHeartbeatMessage received");
                switch (m_LocalState)
                {
                    case State_Master: // Another Master exists (the one with lowest IP address should demote to Slave state)
                        WriteEventLogEntry("M recd M_Heartbeat");
                        CompareNodeIDs_DemoteIfRemoteIsHigher(SenderIPAddress);
                        break;
                    case State_Candidate: // A Master exists (abandon election and demote to Slave state)
                        WriteEventLogEntry("C recd M_Heartbeat");
                        TransitionToSlaveState();
                        break;
                    case State_Slave: // A Master exists (no action needed)
                        WriteEventLogEntry("S recd M_Heartbeat");
                        m_iMostRecentHeartbeat_TimeDetected = m_iMasterHeartbeatClock; 	// Update MostRecentHeartbeat timestamp
                        break;
                }
                break;
            case MT_ElectionTrigger:
                WriteEventLogEntry("ElectionTriggerMessage received");
                switch (m_LocalState)
                {
                    case State_Master: // Another node thinks the Master does not exist - send heartbeat immediately
                        WriteEventLogEntry("M recd Election");
                        SendMasterHeartbeatMessage();
                        break;
                    case State_Candidate: // This is the 'election proper' (the candidate with lowest IP address should demote to Slave state)
                        WriteEventLogEntry("C recd Election");
                        CompareNodeIDs_DemoteIfRemoteIsHigher(SenderIPAddress);
                        if (BULLY_ELECTION_ALGORITHM_Process_State.State_Candidate == m_LocalState)
                        {	// Still in Candidate state, so won the challenge
                            SendElectionTriggerMessage(); // Signal the local node's ID to other nodes (do this every time we are challenged)
                            m_iClock_CountInCandidateState_SincelastChallenge = 0; // Reset Candidate duration clock
                            WriteEventLogEntry("Local won");
                        }
                        break;
                    case State_Slave:	// This is the 'election proper' (if the local ID is higher than the sender's then we join the election)
                        WriteEventLogEntry("S recd Election");
                        CompareNodeIDs_PromoteToCandidateIfRemoteIsLower(SenderIPAddress);
                        if (BULLY_ELECTION_ALGORITHM_Process_State.State_Candidate == m_LocalState)
                        {	// Now in Candidate state, so won the challenge
                            WriteEventLogEntry("Local won");
                        }
                        break;
                }
                break;
        }
    }

    void CompareNodeIDs_DemoteIfRemoteIsHigher(InetAddress SenderIPAddress)
    {
        if (false == CompareNodeIDs_RemoteIsLower(SenderIPAddress))
        {   // Remote has higher address - the local node looses and must demote to Slave state
            TransitionToSlaveState();
        }
    }

    void CompareNodeIDs_PromoteToCandidateIfRemoteIsLower(InetAddress SenderIPAddress)
    {
        if (true == CompareNodeIDs_RemoteIsLower(SenderIPAddress))
        {   // Remote has lower address - the local node wins and must promote to Candidate state
            TransitionToCandidateState();
        }
    }

    Boolean CompareNodeIDs_RemoteIsLower(InetAddress SenderIPAddress)
    {
        byte[] bLocalAd = m_Local_IPAddress.getAddress();

        int iLocal = IPAddressByteArray_ToInt(bLocalAd);

        byte[] bSenderAd = SenderIPAddress.getAddress();

        int iSender = IPAddressByteArray_ToInt(bSenderAd);

        if (iLocal > iSender)
        {   // Local node has the higher ID (its IP address)
            return true;
        }
        // Note, since only one process can exist at each computer (due to port binding), the case where both addresses are equal cannot occur.
        return false; // Local node has the lower ID (its IP address)
    }

    Boolean CompareNodeIDs_SenderSameAsLocal(InetAddress SenderIPAddress)
    {
        return m_Local_IPAddress.equals(SenderIPAddress);
    }

    void ChangeState(BULLY_ELECTION_ALGORITHM_Process_State NewState)
    {
        String sPrevious = ConvertLocalStateToString(); // Remember previous state for display update
        m_LocalState = NewState;                        // Do the state update
        String sNew = ConvertLocalStateToString();	
        String sDisplay;
        sDisplay = String.format("StateChange %s to %s",sPrevious,sNew);
        WriteEventLogEntry(sDisplay);
        DisplayLocalState();
    }

    void DisplayLocalState()
    {
        jTextField_CurrentState.setText(ConvertLocalStateToString());
    }

    String ConvertLocalStateToString()
    {
        switch (m_LocalState)
        {
            case State_Master:
                return "Master";
            case State_Candidate:
                return "Candidate";
            case State_Slave:
                return "Slave";
        }
        return "Unknown";
    }

    void CloseSocketsAndExit()
    {
        Stop_MasterHeartbeatTimer();
        Stop_MessageReceiveTimer();
        try
        {
            m_RecvSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not created
        }
        try
        {
            m_SendSock.close();
        }
        catch (Exception Ex)
        {   // Generic approach to dealing with situations such as socket not created
        }
        System.exit(0);
    }
        
    private void WriteEventLogEntry(String sStr)
    {
        sStr += "\n";
        m_listModel_EventLog.addElement(sStr);       
    }
        
    public int IPAddressByteArray_ToInt(byte[] bVal)
    {
        if(4 != bVal.length)
        {
            return -1;  // Signal invalid byte array length
        }
        int iResult = bVal[3] & 0xFF;
        iResult += ((bVal[2] & 0xFF) << 8);
        iResult += ((bVal[1] & 0xFF) << 16);
        iResult += ((bVal[0] & 0xFF) << 24); 
        return iResult;
    }
    
    InetAddress GetLocalBroadcastAddress_From_LocalHostAddress(InetAddress LocalAddress) throws Exception
    {   // Assumes IPv4 address
        try
        {
            byte[] bLocalBroadcast = LocalAddress.getAddress(); // Initialise byte array with local host's address (e.g. 192.168.100.6)
            // Set the host portion of the address to the broadcast value, depending on the address class
            // Class A from 0.0.0.0 to 127.255.255.255   will be converted to the broadcast address x.255.255.255 
            // Class B from 128.0.0.0 to 191.255.255.255   will be converted to the broadcast address x.x.255.255  
            // Class C 192.0.0.0 to 223.255.255.255   will be converted to the broadcast address x.x.x.255 
            // Class D and E (224.0.0.0 to 225.255.255.254)   will be converted to the broadcast address 255.255.255.255 
            
            // For comparison purposes, the first byte of the address is converted to an int value (to ensure operations are performed 'unsigned')
            int iMostSignificantAddrsssByte = (int) bLocalBroadcast[0] & 0xFF;
            if(iMostSignificantAddrsssByte >= 0 && iMostSignificantAddrsssByte <= 127)
            {   // Class A address
                bLocalBroadcast[1] = (byte)255;
                bLocalBroadcast[2] = (byte)255;
                bLocalBroadcast[3] = (byte)255;
            }
            if(iMostSignificantAddrsssByte >= 128 && iMostSignificantAddrsssByte <= 191)
            {   // Class B address
                bLocalBroadcast[2] = (byte)255;
                bLocalBroadcast[3] = (byte)255;
            }
            if(iMostSignificantAddrsssByte >= 192 && iMostSignificantAddrsssByte <= 223)
            {   // Class C address
                bLocalBroadcast[3] = (byte)255;
            }
            if(iMostSignificantAddrsssByte >= 224)
            {   // Class D / E address
                bLocalBroadcast[0] = (byte)255;
                bLocalBroadcast[1] = (byte)255;
                bLocalBroadcast[2] = (byte)255;
                bLocalBroadcast[3] = (byte)255;
            }
            InetAddress LocalBroadcastAddress = InetAddress.getByAddress(bLocalBroadcast);
            return LocalBroadcastAddress;
        }
        catch (Exception Ex)
        {
            throw(Ex);
        }
    }

    public static void main(String args[]) throws Exception
    {   // Initialise the GUI libraries
        try 
        {
            javax.swing.UIManager.LookAndFeelInfo[] installedLookAndFeels=javax.swing.UIManager.getInstalledLookAndFeels();
            for (int idx=0; idx<installedLookAndFeels.length; idx++)
            {
                if ("Nimbus".equals(installedLookAndFeels[idx].getName())) {
                    javax.swing.UIManager.setLookAndFeel(installedLookAndFeels[idx].getClassName());
                    break;
                }
            }
        } 
        catch (Exception Ex)
        {
            System.exit(0);
        }

        // Create and display the GUI form
        java.awt.EventQueue.invokeAndWait(new Runnable()
        { 
            public void run() 
            { 
                new BullyElectionAlgorithm().setVisible(true);
            }
        } );
    }    

    public void actionPerformed(ActionEvent e)
    {
        if(jButton_ClearLog == e.getSource())
        {   // Clear the Event Log text area
            m_listModel_EventLog.clear();
        }
        if(jButton_Done == e.getSource())
        {
            CloseSocketsAndExit();
        }
    }
        
    private JPanel jPanel_Controls;
    private JLabel jLabel_CurrentState;
    private JTextField jTextField_CurrentState;
    private JButton jButton_ClearLog;
    private JButton jButton_Done;
    private JPanel jPanel_EventLog;
    private JList jList_EventLog;
    private JScrollPane jScrollPane_EventLog;

    private void initComponents()
    {
        jLabel_CurrentState = new javax.swing.JLabel();
        jLabel_CurrentState.setText("Current State");
        jTextField_CurrentState = new javax.swing.JTextField();
        jTextField_CurrentState.setMaximumSize(new Dimension(100, 4));
        jTextField_CurrentState.setHorizontalAlignment(JTextField.CENTER);        
        jPanel_Controls = new javax.swing.JPanel();
        jButton_ClearLog = new JButton();
        jButton_ClearLog.setText("Clear Event List");
        jButton_ClearLog.setMaximumSize(new Dimension(30, 4));
        jButton_ClearLog.addActionListener(this);
        jButton_Done = new JButton();
        jButton_Done.setText("Done");
        jButton_Done.setMaximumSize(new Dimension(30, 4));
        jButton_Done.addActionListener(this);
        jPanel_Controls.setPreferredSize(new Dimension(100, 400));
        jPanel_Controls.setBorder(javax.swing.BorderFactory.createTitledBorder("Controls"));

        org.jdesktop.layout.GroupLayout jPanel1Layout = new org.jdesktop.layout.GroupLayout(jPanel_Controls);
        jPanel_Controls.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jLabel_CurrentState)
                .add(jTextField_CurrentState)
                .add(jButton_ClearLog)
                .add(jButton_Done));
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel1Layout.createSequentialGroup()
                    .add(jLabel_CurrentState)
                    .add(jTextField_CurrentState)
                    .addContainerGap(100, 140)
                    .add(jButton_ClearLog)
                    .addContainerGap(100, 140)
                    .add(jButton_Done)));
        jPanel1Layout.linkSize(new java.awt.Component[] {jButton_ClearLog, jButton_Done});

        jPanel_EventLog = new javax.swing.JPanel();
        jList_EventLog = new JList(m_listModel_EventLog);
        jList_EventLog.setMaximumSize(new Dimension(300, 400));
        jList_EventLog.setSelectedIndex(0);
        jList_EventLog.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        jScrollPane_EventLog = new javax.swing.JScrollPane(jList_EventLog,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, 
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        jPanel_EventLog.setPreferredSize(new Dimension(300, 400));
        jPanel_EventLog.setBorder(javax.swing.BorderFactory.createTitledBorder("Event List"));
        
        org.jdesktop.layout.GroupLayout jPanel2Layout = new org.jdesktop.layout.GroupLayout(jPanel_EventLog);
        jPanel_EventLog.setLayout(jPanel2Layout);
        jPanel2Layout.setHorizontalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                .add(jScrollPane_EventLog));
        jPanel2Layout.setVerticalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                .add(jScrollPane_EventLog));        

        org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        getContentPane().setPreferredSize(new Dimension(460, 420));
         
        layout.setHorizontalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(layout.createSequentialGroup()
                    .add(jPanel_Controls)
                    .add(jPanel_EventLog)));

        layout.setVerticalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel_Controls)
                .add(jPanel_EventLog));

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle("Bully Election Algorithm");
        pack();
    }                                             
}