/*	File			Game_Client_DS.java
	Purpose			Network Game (Noughts and Crosses)CLIENT
	Author			Richard John Anthony
	Date			February 2015 (original Cpp version December 2003)
*/
package Game_Client_DS;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.*;
import java.util.Timer;
import java.util.TimerTask;
import javax.swing.ScrollPaneConstants;
import javax.swing.JOptionPane;
import javax.swing.*;
    
public class Game_Client_DS extends javax.swing.JFrame implements ActionListener
{
    private static int PORT = 8000;                 // Default port 8000 decimal (clients connect to the server on this port)
    private static int LOCAL_PORT = 8008;           // Port that ServerAdvertise socket binds to (so not tying up the SERVER_ADVERTISE_PORT,
                                                    // thus allowing a Game client to be co-resident on the same computer as the Server)
    private static int MAX_ALIAS_LEN = 15;
    private static int MAX_SERVER_NAME_LEN = 30;

    private enum MessageType {REGISTER_ALIAS, PLAYER_LIST, CHOOSE_OPPONENT, LOCAL_CELL_SELECTION,
            OPPONENT_CELL_SELECTION, START_OF_GAME, END_OF_GAME, CLIENT_DISCONNECT};

    private enum CodeType {END_OF_GAME_DRAW, END_OF_GAME_WIN, END_OF_GAME_LOSE};

    // To implement the equivallent of a C/C++/C# structure, in Java a class is used
    final class Game_Message_PDU
    {
        MessageType eMessageType;   //size 4                Offset 0
        byte[] cAlias;              //size [15+1]           Offset 4
        byte[] cPlayerList;         //size [10 * (15+1)]    Offset 20   Comma-separated list of aliases
        int iCell;                  //size 4                Offset 180
        CodeType eCode;             //size 4                Offset 184
        int iToken;                 //size 4                Offset 188
        Boolean bFirstMove;         //size 1(padded to 4)   Offset 192
    };

    private static int Game_Message_PDU_Size = 196;
            
    Socket m_ConnectSock;
    InputStream m_InStream;
    OutputStream m_OutStream;
    InetAddress  m_LocalAddress;
    InetAddress  m_GameServerAddress;
    int m_iPort;
    Timer m_Receive_Timer;
    
    Boolean m_bConnected;
    String m_sToken;
    String m_sOpponentToken;
    DefaultListModel m_listModel_AvailablePlayers; // For use with JList

    // Member variable associated with the use of the Directory Service    
    String GAME_SERVER_NAME = "GameServer";	// Game Server name, as registered with DS
     
    public Game_Client_DS()
    {
        m_listModel_AvailablePlayers = new DefaultListModel(); // For use with JList
        initComponents();
        InitialiseControls();
        Invoke_DirectoryService_To_Resolve_Server_Name();
    }

    private void InitialiseControls()    
    {    
        m_bConnected = false;
        jTextField_ContactWithDS.setEnabled(false);
        jTextField_ContactWithDS.setText("Not yet Contacted DS");
        jTextField_ConnectionStatus.setEnabled(false);
        jTextField_ConnectionStatus.setText("NOT Connected to Game Server");
        m_iPort = PORT;
        jTextField_Port.setEnabled(false);
        jTextField_Port.setText(Integer.toString(m_iPort));
        jTextField_IPAddress.setEnabled(false);
        jTextField_PlayerAlias.setEnabled(false);
        jButton_RegisterAlias.setEnabled(false);
        jList_AvailablePlayers.setEnabled(false);
        JScrollPane_AvailablePlayers.setEnabled(true);
        jButton_SelectPlayer.setEnabled(false);    
        jLabel_OpponentAlias.setEnabled(false); 
        jTextField_OpponentAlias.setEnabled(false);       
        ClearAllCells();
        jTextField_GameStatus.setEnabled(false);
        jTextField_GameStatus.setText("No active game"); // Display game status
        jTextField_Token.setEnabled(false);
        jTextField_MoveStatus.setEnabled(false); 
       
        m_LocalAddress = GetLocalIPAddress();
    }

    private InetAddress GetLocalIPAddress()
    {   
        String sHostAddress = "";
        try
        {
            InetAddress LocalAddress = InetAddress.getLocalHost();
            sHostAddress = LocalAddress.toString(); // Expected format: "Hostname/192.168.100.5"
            int iStartOfDottedIPSubstring = sHostAddress.indexOf('/');
            if(-1 == iStartOfDottedIPSubstring)
            {   // Align to the actual start of the dotted IP address part of the string
                iStartOfDottedIPSubstring = 0; // '/' not found
            }
            else
            {
                iStartOfDottedIPSubstring++; // Skip the '/' charachter
            }
            String sDottedIPAddress = sHostAddress.substring(iStartOfDottedIPSubstring);
            jTextField_IPAddress.setText(sDottedIPAddress);
            return LocalAddress;
        }
        catch (Exception Ex)
        {
            JOptionPane.showMessageDialog(null,Ex.toString(),"Noughts Crosses CLIENT",JOptionPane.PLAIN_MESSAGE);
            CloseAndExit();
        }
        return null;
    }
    
    void Invoke_DirectoryService_To_Resolve_Server_Name()
    {        
        DirectoryServiceClient m_DirectoryServiceClient;
        jTextField_ContactWithDS.setText("Contacted DS");
        m_DirectoryServiceClient = new DirectoryServiceClient();
        m_DirectoryServiceClient.Resolve_Name_to_IP(GAME_SERVER_NAME, true);
        jTextField_ContactWithDS.setText("DS resolved server name");
        m_GameServerAddress = m_DirectoryServiceClient.Get_ResolvedAddress();
        m_iPort = m_DirectoryServiceClient.Get_Port();
        jTextField_Port.setText(Integer.toString(m_iPort));
        String sAddress = m_GameServerAddress.toString();
        jTextField_IPAddress.setText(sAddress);
        if(0 == m_iPort) // DS returns 0 value in address and port if requested name was not known
        {           
            JOptionPane.showMessageDialog(null,"Game Server not registered with Directory Service","Noughts Crosses CLIENT",JOptionPane.PLAIN_MESSAGE);
            CloseAndExit();
        }
        Connect_to_Server();
    }

    private void CreateAndConnectSocket()
    {
        try
        {
            m_ConnectSock = new Socket(m_GameServerAddress, m_iPort); // Create Socket and connect to Game Server
            m_InStream = m_ConnectSock.getInputStream();
            m_OutStream = m_ConnectSock.getOutputStream();
            m_ConnectSock.setSoTimeout(0); // 0 indicates infinite timeout (i.e blocking IO behaviour)
            jTextField_ConnectionStatus.setText("CONNECTED to server");
        }  
        catch (Exception Ex)  // Catch exceptions from the socket operations
        {
           JOptionPane.showMessageDialog(null,Ex.toString(),"Noughts Crosses CLIENT",JOptionPane.PLAIN_MESSAGE);
           CloseAndExit();
        }
    }

    private void Connect_to_Server()
    {	
        CreateAndConnectSocket();
        m_bConnected = true;
        jTextField_Port.setEnabled(false);
        jTextField_IPAddress.setEnabled(false);
        jTextField_PlayerAlias.setEnabled(true);
        jButton_RegisterAlias.setEnabled(true);
        Start_Receive_Timer(); // Enable receiving (check periodically)
    }

    private void Start_Receive_Timer()
    {
        m_Receive_Timer = new Timer();
        m_Receive_Timer.scheduleAtFixedRate(
                new OnTimedEvent_Receive_Timer(), 100, 100); // Short interval to allow UI responsiveness)
    }

    private void StopTimer()
    {
        if (null != m_Receive_Timer)
        {
            m_Receive_Timer.cancel();
        }
    }

    private class OnTimedEvent_Receive_Timer extends TimerTask
    {
        public void run()
        {
            if (true == m_bConnected)
            {
                DoReceive();
            }
        }
    }
    
    private void DoReceive()
    {		
        byte[] bBuffer = new byte[1024];  // Create an appropriately-sized buffer to Receive into
        int iReceiveByteCount;
        try
        {
            iReceiveByteCount = m_InStream.read(bBuffer);
        }
        catch (SocketTimeoutException Ex)
        {
            return;
        }
        catch (Exception Ex)
        {
            return;
        }
        if(-1 == iReceiveByteCount)
        {
            return; // no data
        }
        if(Game_Message_PDU_Size <= iReceiveByteCount)
        {
            // Deserialise the Receive buffer to get the Game message structure
            Game_Message_PDU Message = DeSerialize_Game_Message_PDU(bBuffer);
            switch (Message.eMessageType)
            {
                case PLAYER_LIST:
                    DisplayPlayerList(Message);	
                    break;
                case START_OF_GAME:
                    ClearAllCells();
                    jButton_SelectPlayer.setEnabled(false);             // Disable Select Opponent control
                    String sOpponent = new String(Message.cAlias);      // Display opponent alias
                    jTextField_OpponentAlias.setText(sOpponent);
                    
                    if(Message.iToken == 0)                             // Set and display our token
                    {
                        m_sToken = "    0    ";
                        m_sOpponentToken = "    X    ";
                    }
                    else
                    {
                        m_sToken = "    X    ";
                        m_sOpponentToken = "    0    ";
                    }		
                    jTextField_Token.setText(m_sToken);

                    if(Message.bFirstMove == true)                      // Configure first-move logic
                    {
                        jTextField_MoveStatus.setText("Your Move");
                        EnableEmptyCells();
                    }
                    else
                    {
                        jTextField_MoveStatus.setText("Opponent's Move");
                        DisableEmptyCells();
                    }
                    jTextField_GameStatus.setText("Game In Progress");  // Display game status
                    break;
                case END_OF_GAME:
                    switch(Message.eCode) 
                    {
                        case END_OF_GAME_DRAW:
			    JOptionPane.showMessageDialog(null,"The game was a draw","Game Over",JOptionPane.PLAIN_MESSAGE);
                            break;
                        case END_OF_GAME_WIN:
			    JOptionPane.showMessageDialog(null,"Well done! You won","Game Over",JOptionPane.PLAIN_MESSAGE);
                            break;
                        case END_OF_GAME_LOSE:
			    JOptionPane.showMessageDialog(null,"Unlucky - You lost","Game Over",JOptionPane.PLAIN_MESSAGE);
                            break;
                    }
                    CloseAndExit();
                    break;
                case OPPONENT_CELL_SELECTION:
                    jButton_GameCellArray[Message.iCell].setText(m_sOpponentToken);
                    jButton_GameCellArray[Message.iCell].setEnabled(false);
                    jTextField_MoveStatus.setText("Your Move");
                    EnableEmptyCells();
                default:
                break;
	    }
        }
    }

    private void DisplayPlayerList(Game_Message_PDU Message)
    {
        m_listModel_AvailablePlayers.clear();
        String sAlias = jTextField_PlayerAlias.getText(); // Read our Alias (exclude it from displayed list)
        String sPlayerList = TruncateString(new String(Message.cPlayerList),160);
        int iLength = sPlayerList.length();
        // sPlayerList is a comma-separated, ASCII (single byte per char) null-terminated list (terminated at iLength)
        // There is no comma after the last item in the list
        int iNextCommaPos;
            String sCurrentPlayer = "";
        while ((iNextCommaPos = sPlayerList.indexOf(',')) > 0 && iNextCommaPos < iLength) // Ignore chars beyond the valid string content
        {
            sCurrentPlayer = sPlayerList.substring(0/*start*/, iNextCommaPos/*length*/);
            if (!sCurrentPlayer.equalsIgnoreCase(sAlias)) // Exclude own alias from displayed list
            {
                m_listModel_AvailablePlayers.addElement(sCurrentPlayer);
            }
                sPlayerList = sPlayerList.substring(iNextCommaPos + 1/*start*/);// Remove 'dealt with' name from head of list
                iLength -= (iNextCommaPos + 1);                                 // Keep track of remaining list-string length
            }
        // Deal with residual substring (there is no comma after the last item in the list)
        sPlayerList = TruncateString(sPlayerList, iLength);
        if (0 < iLength && !sPlayerList.equalsIgnoreCase(sAlias)) // Exclude own alias from displayed list
        {
             m_listModel_AvailablePlayers.addElement(sPlayerList);
        }

        if (0 >= m_listModel_AvailablePlayers.getSize())
        {
            jList_AvailablePlayers.setEnabled(false);
            jButton_SelectPlayer.setEnabled(false);
        }
        else
        {
            jList_AvailablePlayers.setSelectedIndex(0); // Set default selection
            jList_AvailablePlayers.setEnabled(true);
            jButton_SelectPlayer.setEnabled(true);
        }
    }
       
    private void CellSelectAction(int iCell)
    {   // send a CELL_SELECTION message to server
        Game_Message_PDU Message = Message_PDU_Initialise(); 
        Message.eMessageType = MessageType.LOCAL_CELL_SELECTION;
        Message.iCell = iCell;
        SendMessage(Message);
        jTextField_MoveStatus.setText("Opponent's Move");
        DisableEmptyCells();
    }

    private void DisableEmptyCells()
    {
        for(int iCellIndex = 0; iCellIndex < 9; iCellIndex++)
        {   // Test each button in turn, to determine if its displayed text is blank or contains a 'X' or '0'
            String sButtonText = jButton_GameCellArray[iCellIndex].getText();
            if(-1 == sButtonText.indexOf('X') && -1 == sButtonText.indexOf('0'))
            {   // Button is blank - disable it
                jButton_GameCellArray[iCellIndex].setEnabled(false);
            }
        }
    }

    private void EnableEmptyCells()
    {
        for(int iCellIndex = 0; iCellIndex < 9; iCellIndex++)
        {   // Test each button in turn, to determine if its displayed text is blank or contains a 'X' or '0'
            String sButtonText = jButton_GameCellArray[iCellIndex].getText();
            if(-1 == sButtonText.indexOf('X') && -1 == sButtonText.indexOf('0'))
            {   // Button is blank - enable it
                jButton_GameCellArray[iCellIndex].setEnabled(true);
            }
        }
    }

    private void ClearAllCells()
    {
        for(int iCellIndex = 0; iCellIndex < 9; iCellIndex++)
        {
            jButton_GameCellArray[iCellIndex].setEnabled(false);
            jButton_GameCellArray[iCellIndex].setText("          ");
        }
    }

    private void SendMessage(Game_Message_PDU Message)
    {
        byte[] bBuffer = Serialize_Game_Message_PDU(Message);
        try
        {
            m_OutStream.write(bBuffer);
        }
        catch (SocketTimeoutException Ex)
        {
        }
        catch (Exception ex)
        {
        }
    }
    
    private void CloseConnectionAndSocket()
    {
        try
        {
             m_ConnectSock.close();    // Close the socket
        }
        catch (Exception Ex)
        {
        }
    }

    private void CloseAndExit()
    {   // Close all sockets and exit the application
        StopTimer();
        CloseConnectionAndSocket();
        System.exit(0);
    }
        
    private Game_Message_PDU Message_PDU_Initialise()
    {
        Game_Message_PDU M = new Game_Message_PDU();
        M.eMessageType = MessageType.END_OF_GAME; // This field is always overwritten as used in every instance
        M.cAlias = new byte[MAX_ALIAS_LEN + 1];
        M.cPlayerList = new byte[10 * (MAX_ALIAS_LEN + 1)];
        M.iCell = 0;
        M.eCode = CodeType.END_OF_GAME_DRAW;
        M.iToken = 0;
        M.bFirstMove = false;
        return M;
    }
        
    private static byte[] IntToByteArray(int iVal)
    {
        byte[] bArray = new byte[4];
        bArray[0] = (byte)(iVal & 0x000000ff);
        bArray[1] = (byte)((iVal & 0x0000ff00) >> 8);
        bArray[2] = (byte)((iVal & 0x00ff0000) >> 16);
        bArray[3] = (byte)((iVal & 0xff000000) >> 24);
        return bArray;
    }

    private int ByteArraySize4_ToInt(byte[] bVal)
    {
        if(4 != bVal.length)
        {
            return -1;  // Signal invalid byte array length
        }
        int iResult = bVal[0] & 0xFF;
        iResult += ((bVal[1] & 0xFF) << 8);
        iResult += ((bVal[2] & 0xFF) << 16);
        iResult += ((bVal[3] & 0xFF) << 24);            
        return iResult;
    }
    
    private String ByteArraySize4_To_IPAddessString(byte[] bVal)
    {
        if(4 != bVal.length)
        {
            return "";  // Signal invalid byte array length
        }
        int iAddr1 = bVal[0] & 0xFF;
        int iAddr2 = bVal[1] & 0xFF;
        int iAddr3 = bVal[2] & 0xFF;
        int iAddr4 = bVal[3] & 0xFF;
        String sAddress = String.format("%d.%d.%d.%d", iAddr1, iAddr2, iAddr3, iAddr4);
        return sAddress;
    }

    private byte[] DottedIPAddessString_To_ByteArraySize4(String sDottedIPAddress)
    {
        byte[] bAddress = new byte[4]; // Each byte initialised as value 0
        int iLength = sDottedIPAddress.length();
        if(7 > iLength || 15 < iLength) // A well formed dotted IP address string must be in this length range
        {
            return bAddress; // All 0's signals mal-formed dotted IPaddress string
        }
        // sDottedIPAddress is a '.' separated, list of numeric values
        // There is no '.' after the last item in the list
        int iNextDotPos  = 0;
        String sCurrentValue = "";
        int iArrayIndex = 0;
        try
        {
            while ((iNextDotPos = sDottedIPAddress.indexOf('.')) > 0 && iNextDotPos < iLength) // Ignore chars beyond the valid string content
            {
                sCurrentValue = sDottedIPAddress.substring(0/*start*/, iNextDotPos/*length*/);
                bAddress[iArrayIndex++] = (byte)Integer.parseInt(sCurrentValue);
                sDottedIPAddress = sDottedIPAddress.substring(iNextDotPos + 1/*start*/);// Remove 'dealt with' name from head of list
                iLength -= (iNextDotPos + 1);                                 // Keep track of remaining list-string length
            }
            // Deal with residual substring (there is no comma after the last item in the list)
            //sDottedIPAddress = TruncateString(sDottedIPAddress, iLength);
            bAddress[iArrayIndex++] = (byte)Integer.parseInt(sDottedIPAddress);
        }
        catch (Exception Ex)
        {   // Catch various string-format related exceptions (such as having more than three '.'s which
            // would cause the byte array index to go out of bounds)
            bAddress[0] = 0; // signal mal-formed dotted IPaddress string 
        }
        return bAddress;
    }

    private String TruncateString(String sStr, int iMaxLength)
    {
        try
        {
            // Truncate to null terminator if string contains null-termiation chars (happens when converting from a byte array)
            int iNullTerminatePosition = sStr.indexOf(0);
            if(-1 != iNullTerminatePosition)
            {
                sStr = sStr.substring(0, iNullTerminatePosition);
            }
            if(sStr.length()> iMaxLength)
            {   // Only use substring if the string is actually longer than the second parameter
                // (the length limit) otherwise an ArgumentOutOfRangeException is thrown
                sStr = sStr.substring(0, iMaxLength);
            }
            return sStr;
        }
        catch (Exception Ex)
        {
            return "";
        }
    }
        
    private byte[] Serialize_Game_Message_PDU(Game_Message_PDU Message)
    {   // Convert an Message_PDU structure into a byte array
        byte[] bByteArray = new byte[Game_Message_PDU_Size];    // Create a byte array to hold the serialised form of the message PDU
        bByteArray[0] = (byte)Message.eMessageType.ordinal();
        System.arraycopy(Message.cAlias/*src*/, 0, bByteArray/*dest*/, 4, Message.cAlias.length/*count*/);
        System.arraycopy(Message.cPlayerList/*src*/, 0, bByteArray/*dest*/, 20, Message.cPlayerList.length/*count*/);
        byte[] bInt = IntToByteArray(Message.iCell);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 180, 4/*count*/); // Next 4-byte boundary
        bByteArray[184] = (byte)Message.eCode.ordinal();
        bInt = IntToByteArray(Message.iToken);
        System.arraycopy(bInt/*src*/, 0, bByteArray/*dest*/, 188, 4/*count*/); // Next 4-byte boundary
    
        bByteArray[192] = (byte)0;  // Boolean serialises to 1 byte
        if(true == Message.bFirstMove)
        {   
            bByteArray[192] = (byte)1;
        }
        return bByteArray;
    }

    private Game_Message_PDU DeSerialize_Game_Message_PDU(byte[] bRecdByteArray)
    {   // Converts a byte array (from Receive) to a EventNotificationService_Message_PDU structure (Fields have fixed length, but non-fixed data length)
        Game_Message_PDU Game_PDU = new Game_Message_PDU();
        Game_PDU.eMessageType = MessageType.values()[bRecdByteArray[0]]; // Convert byte value to enumerated MessageType
        Game_PDU.cAlias = new byte[MAX_ALIAS_LEN+1];
        System.arraycopy(bRecdByteArray/*src*/, 4, Game_PDU.cAlias/*dest*/, 0, MAX_ALIAS_LEN/*count*/); 
        Game_PDU.cPlayerList = new byte[10 * (MAX_ALIAS_LEN+1)];
        System.arraycopy(bRecdByteArray/*src*/, 20, Game_PDU.cPlayerList/*dest*/, 0, (10 * (MAX_ALIAS_LEN+1))-1/*count*/); 
        byte[] bInt = new byte[4]; 
        System.arraycopy(bRecdByteArray/*src*/, 180, bInt/*dest*/, 0, 4/*count*/); 
        Game_PDU.iCell = ByteArraySize4_ToInt(bInt);            
        Game_PDU.eCode = CodeType.values()[bRecdByteArray[184]]; // Convert byte value to enumerated MessageType
        System.arraycopy(bRecdByteArray/*src*/, 188, bInt/*dest*/, 0, 4/*count*/); 
        Game_PDU.iToken = ByteArraySize4_ToInt(bInt); 
        Game_PDU.bFirstMove = true;
        if(0 == bRecdByteArray[192])
        {
            Game_PDU.bFirstMove = false;
        }
        return Game_PDU;
    }
    
    public static void main(String args[]) throws Exception
    {
        // Initialise the GUI libraries
        try 
        {
            javax.swing.UIManager.LookAndFeelInfo[] installedLookAndFeels=javax.swing.UIManager.getInstalledLookAndFeels();
            for (int idx=0; idx<installedLookAndFeels.length; idx++)
            {
                if ("Nimbus".equals(installedLookAndFeels[idx].getName())) {
                    javax.swing.UIManager.setLookAndFeel(installedLookAndFeels[idx].getClassName());
                    break;
                }
            }
        } 
        catch (Exception Ex)
        {
            System.exit(0);
        }

        // Create and display the GUI form
        java.awt.EventQueue.invokeAndWait(new Runnable()
        { 
            public void run() 
            { 
                new Game_Client_DS().setVisible(true);
            }
        } );
    }    

    public void actionPerformed(ActionEvent e)
    {
        if(jButton_Done == e.getSource())
        {
            CloseAndExit();
        }
        for(int iCellIndex = 0; iCellIndex < 9; iCellIndex++)
        {
            if(jButton_GameCellArray[iCellIndex] == e.getSource())
            {
 	        jButton_GameCellArray[iCellIndex].setText(m_sToken);
	        jButton_GameCellArray[iCellIndex].setEnabled(false);
	        CellSelectAction(iCellIndex);
            }
        }
        if(jButton_RegisterAlias == e.getSource())
        {
            String sAlias = jTextField_PlayerAlias.getText();
            if(0 == sAlias.length())
            {
                JOptionPane.showMessageDialog(null,"Alias name cannot be blank", "'Noughts and Crosses' Client",JOptionPane.PLAIN_MESSAGE);              
            }
            else
            {
                Game_Message_PDU Message = Message_PDU_Initialise();
                Message.eMessageType = MessageType.REGISTER_ALIAS;
                sAlias = TruncateString(sAlias, MAX_ALIAS_LEN); // Truncate if necessary
                jTextField_PlayerAlias.setText(sAlias); // Update textbox with (possibly) truncated string
                Message.cAlias = sAlias.getBytes();
                SendMessage(Message);
                jTextField_PlayerAlias.setEnabled(false);	
                jButton_RegisterAlias.setEnabled(false);
            }
        }
        if(jButton_SelectPlayer == e.getSource())
        {
            int iSelection = jList_AvailablePlayers.getSelectedIndex();
            if(-1 == iSelection)
            {
                JOptionPane.showMessageDialog(null,"No player selected", "'Noughts and Crosses' Client",JOptionPane.PLAIN_MESSAGE);              
            }
            else
            {
                // Send a CHOOSE_OPPONENT message to server
                Game_Message_PDU Message = Message_PDU_Initialise(); 
                Message.eMessageType = MessageType.CHOOSE_OPPONENT;
                Message.cAlias = jList_AvailablePlayers.getSelectedValue().toString().getBytes();
                SendMessage(Message);
            }
        }
    }
    
    private JPanel jPanel_Communication;
    private JTextField jTextField_ContactWithDS;
    private JTextField jTextField_ConnectionStatus;
    private JLabel jLabel_Port;
    private JTextField jTextField_Port;
    private JLabel jLabel_IPAddress;
    private JTextField jTextField_IPAddress;
    private JPanel jPanel_PlayerAlias;
    private JTextField jTextField_PlayerAlias;
    private JButton jButton_RegisterAlias;
    private JPanel jPanel_AvailablePlayers;
    private JList jList_AvailablePlayers;
    private JScrollPane JScrollPane_AvailablePlayers;
    private JButton jButton_SelectPlayer;
    private JLabel jLabel_OpponentAlias;
    private JTextField jTextField_OpponentAlias;
    private JPanel jPanel_Game;
    private JButton[] jButton_GameCellArray;
    private JLabel jLabel_GameStatus;
    private JTextField jTextField_GameStatus;
    private JLabel jLabel_Token;
    private JTextField jTextField_Token;
    private JLabel jLabel_MoveStatus;
    private JTextField jTextField_MoveStatus;
    private javax.swing.JButton jButton_Done;

    private void initComponents()
    {
        jTextField_ContactWithDS = new JTextField();
        jTextField_ContactWithDS.setMaximumSize(new Dimension(200, 4));
        jTextField_ContactWithDS.setHorizontalAlignment(JTextField.CENTER);
        jTextField_ConnectionStatus = new JTextField();
        jTextField_ConnectionStatus.setMaximumSize(new Dimension(200, 4));
        jTextField_ConnectionStatus.setHorizontalAlignment(JTextField.CENTER);
        jLabel_Port = new JLabel();
        jLabel_Port.setText("Port");
        jTextField_Port = new JTextField();
        jTextField_Port.setMaximumSize(new Dimension(100, 4));
        jTextField_Port.setHorizontalAlignment(JTextField.CENTER);
        jLabel_IPAddress = new JLabel();
        jLabel_IPAddress.setText("IPv4 Address");
        jTextField_IPAddress = new JTextField();
        jTextField_IPAddress.setMaximumSize(new Dimension(100, 4));
        jTextField_IPAddress.setHorizontalAlignment(JTextField.CENTER);
        jPanel_Communication = new JPanel();
        jPanel_Communication.setPreferredSize(new Dimension(200, 400));
        jPanel_Communication.setBorder(BorderFactory.createTitledBorder("Communication configuration"));

        org.jdesktop.layout.GroupLayout jPanel1Layout = new org.jdesktop.layout.GroupLayout(jPanel_Communication);
        jPanel_Communication.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jPanel1Layout.createSequentialGroup()
                .add(jTextField_ContactWithDS)           
                .add(jTextField_ConnectionStatus))
            .add(jPanel1Layout.createSequentialGroup()
                .add(jLabel_Port)
                .addContainerGap(20, 20)
                .add(jTextField_Port)
                .addContainerGap(20, 20)
                .add(jLabel_IPAddress)
                .addContainerGap(20, 20)
                .add(jTextField_IPAddress)));
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jPanel1Layout.createSequentialGroup()
                .add(jPanel1Layout.createParallelGroup()
                    .add(jTextField_ContactWithDS)
                    .add(jTextField_ConnectionStatus))
                .addContainerGap(20, 20)
                .add(jPanel1Layout.createParallelGroup()
                    .add(jLabel_Port)
                    .add(jTextField_Port)
                    .add(jLabel_IPAddress)
                    .add(jTextField_IPAddress))));            
        
        jTextField_PlayerAlias = new JTextField();
        jTextField_PlayerAlias.setMaximumSize(new Dimension(100, 4));
        jTextField_PlayerAlias.setHorizontalAlignment(JTextField.CENTER);
        jButton_RegisterAlias = new JButton();
        jButton_RegisterAlias.setText("Register Alias with Server");
        jButton_RegisterAlias.setMaximumSize(new Dimension(90, 6));
        jButton_RegisterAlias.addActionListener(this);
        jPanel_PlayerAlias = new JPanel();
        jPanel_PlayerAlias.setPreferredSize(new Dimension(200, 400));
        jPanel_PlayerAlias.setBorder(BorderFactory.createTitledBorder(" Player Alias"));

        org.jdesktop.layout.GroupLayout jPanel2Layout = new org.jdesktop.layout.GroupLayout(jPanel_PlayerAlias);
        jPanel_PlayerAlias.setLayout(jPanel2Layout);
        jPanel2Layout.setHorizontalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                .add(jPanel2Layout.createSequentialGroup()
                    .add(jTextField_PlayerAlias)
                    .addContainerGap(10, 10)
                    .add(jButton_RegisterAlias)));
        jPanel2Layout.setVerticalGroup(
            jPanel2Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jTextField_PlayerAlias)
                .add(jButton_RegisterAlias));

        jList_AvailablePlayers = new JList(m_listModel_AvailablePlayers);
        jList_AvailablePlayers.setMaximumSize(new Dimension(100, 130));
        jList_AvailablePlayers.setSelectedIndex(0);
        jList_AvailablePlayers.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        JScrollPane_AvailablePlayers = new javax.swing.JScrollPane(jList_AvailablePlayers,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, 
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        jButton_SelectPlayer = new JButton();
        jButton_SelectPlayer.setText("Select Highlighted Player");
        jButton_SelectPlayer.setMaximumSize(new Dimension(90, 6));
        jButton_SelectPlayer.addActionListener(this);
        jLabel_OpponentAlias = new JLabel();
        jLabel_OpponentAlias.setText("Opponent Alias");
        jTextField_OpponentAlias = new JTextField();
        jTextField_OpponentAlias.setMaximumSize(new Dimension(100, 4));
        jTextField_OpponentAlias.setHorizontalAlignment(JTextField.CENTER);
        jPanel_AvailablePlayers = new JPanel();
        jPanel_AvailablePlayers.setPreferredSize(new Dimension(200, 400));
        jPanel_AvailablePlayers.setBorder(BorderFactory.createTitledBorder("Available Players"));

        org.jdesktop.layout.GroupLayout jPanel3Layout = new org.jdesktop.layout.GroupLayout(jPanel_AvailablePlayers);
        jPanel_AvailablePlayers.setLayout(jPanel3Layout);
        jPanel3Layout.setHorizontalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(jPanel3Layout.createSequentialGroup()
                    .add(JScrollPane_AvailablePlayers)    
                    .addContainerGap(20, 20)
                .add(jPanel3Layout.createParallelGroup()
                    .add(jButton_SelectPlayer)
                    .add(jLabel_OpponentAlias)
                    .add(jTextField_OpponentAlias))));
        jPanel3Layout.setVerticalGroup(
            jPanel3Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
                .add(JScrollPane_AvailablePlayers)
                .add(jPanel3Layout.createSequentialGroup()
                    .add(jButton_SelectPlayer)
                    .addContainerGap(20, 20)
                    .add(jLabel_OpponentAlias)
                    .add(jTextField_OpponentAlias)));
    
        jPanel_Game = new JPanel();
        jPanel_Game.setPreferredSize(new Dimension(200, 400));
        jPanel_Game.setBorder(BorderFactory.createTitledBorder("Game action and state"));
        jButton_GameCellArray = new JButton[9];
        String sStr = "";
        for(int iCellIndex = 0; iCellIndex < 9; iCellIndex++)
        {
            jButton_GameCellArray[iCellIndex] = new JButton();
            jButton_GameCellArray[iCellIndex].setText("          ");
            jButton_GameCellArray[iCellIndex].setMaximumSize(new Dimension(30, 30));
            jButton_GameCellArray[iCellIndex].addActionListener(this);
        }
        jLabel_GameStatus = new JLabel();
        jLabel_GameStatus.setText("Game Status");
        jTextField_GameStatus = new JTextField();
        jTextField_GameStatus.setMaximumSize(new Dimension(130, 4));
        jTextField_GameStatus.setHorizontalAlignment(JTextField.CENTER);
        jLabel_Token = new JLabel();
        jLabel_Token.setText("Your token is");
        jTextField_Token = new JTextField();
        jTextField_Token.setMaximumSize(new Dimension(30, 4));
        jTextField_Token.setHorizontalAlignment(JTextField.CENTER);
        jLabel_MoveStatus = new JLabel();
        jLabel_MoveStatus.setText("Move Status");        
        jTextField_MoveStatus = new JTextField();
        jTextField_MoveStatus.setMaximumSize(new Dimension(130, 4));
        jTextField_MoveStatus.setHorizontalAlignment(JTextField.CENTER);
        jButton_Done = new JButton();
        jButton_Done.setText("Done");
        jButton_Done.setMaximumSize(new Dimension(100, 4));
        jButton_Done.addActionListener(this);

        org.jdesktop.layout.GroupLayout jPanel4Layout = new org.jdesktop.layout.GroupLayout(jPanel_Game);
        jPanel_Game.setLayout(jPanel4Layout);
        jPanel4Layout.setHorizontalGroup(
            jPanel4Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.LEADING)
                .add(jPanel4Layout.createSequentialGroup()
                    .add(jPanel4Layout.createParallelGroup()
                        .add(jPanel4Layout.createSequentialGroup()
                            .add(jButton_GameCellArray[0])
                            .add(jButton_GameCellArray[1])
                            .add(jButton_GameCellArray[2])) 
                        .add(jPanel4Layout.createSequentialGroup()
                            .add(jButton_GameCellArray[3])
                            .add(jButton_GameCellArray[4])
                            .add(jButton_GameCellArray[5])) 
                        .add(jPanel4Layout.createSequentialGroup()
                            .add(jButton_GameCellArray[6])
                            .add(jButton_GameCellArray[7])
                            .add(jButton_GameCellArray[8]))) 
                    .addContainerGap(20, 20)
                    .add(jPanel4Layout.createParallelGroup()
                        .add(jLabel_GameStatus)
                        .add(jTextField_GameStatus)
                        .add(jPanel4Layout.createSequentialGroup()
                            .add(jLabel_Token)
                            .add(jTextField_Token))
                        .add(jLabel_MoveStatus)
                        .add(jTextField_MoveStatus)
                        .add(jButton_Done))));
        jPanel4Layout.setVerticalGroup(
            jPanel4Layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(jPanel4Layout.createSequentialGroup()
                .add(jPanel4Layout.createParallelGroup()
                    .add(jButton_GameCellArray[0])
                    .add(jButton_GameCellArray[1])
                    .add(jButton_GameCellArray[2])) 
                .add(jPanel4Layout.createParallelGroup()
                    .add(jButton_GameCellArray[3])
                    .add(jButton_GameCellArray[4])
                    .add(jButton_GameCellArray[5])) 
                .add(jPanel4Layout.createParallelGroup()
                    .add(jButton_GameCellArray[6])
                    .add(jButton_GameCellArray[7])
                    .add(jButton_GameCellArray[8]))) 
            .add(jPanel4Layout.createSequentialGroup()
                .add(jLabel_GameStatus)
                .add(jTextField_GameStatus)
                .addContainerGap(10, 10)
                .add(jPanel4Layout.createParallelGroup()
                    .add(jLabel_Token)
                    .add(jTextField_Token))
                .addContainerGap(10, 10)
                .add(jLabel_MoveStatus)
                .add(jTextField_MoveStatus)
                .addContainerGap(40, 40)
                .add(jButton_Done)));

        org.jdesktop.layout.GroupLayout layout = new org.jdesktop.layout.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        getContentPane().setPreferredSize(new Dimension(360, 500));
         
        layout.setHorizontalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(layout.createParallelGroup()            
                .add(jPanel_Communication)
                .add(jPanel_PlayerAlias)
                .add(jPanel_AvailablePlayers)
                .add(jPanel_Game)));
        layout.setVerticalGroup(
            layout.createParallelGroup(org.jdesktop.layout.GroupLayout.CENTER)
            .add(layout.createSequentialGroup()
                .add(jPanel_Communication)
                .add(jPanel_PlayerAlias)
                .add(jPanel_AvailablePlayers)
                .add(jPanel_Game)));

        setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
        setTitle("CaseStudy Noughts and Crosses CLIENT");
        pack();
    }                                             
}